extends Node3D
class_name CameraLens

signal autofocus_start
signal autofocus_stop
signal focused
signal out_of_focus

var _dirty = true
var _recreate_effect = true
var _new_fov:float = 0.0
var camera:Camera3D

@export_node_path(Camera3D) var camera_path:NodePath:
    set(v):
        camera = get_node(v)
        if not camera:
            camera = get_parent()
        camera_path = v

@onready var autofocus_timer = $AutofocusTimer

@export_node_path(WorldEnvironment) var environment: NodePath:
    set(v):
        environment = v
        _dirty = true
        var n = get_node(v)
        if n:
            world_env = n


@export var distance: float = 20.0:
    set(v):
        _dirty = true
        distance = v
        target_distance = v


@export var aperture: float = 5.6:
    set(v):
        _dirty = true
        _aperture_changed = true
        aperture = v

enum ApertureChange {Immediate, Zoom}
@export_enum("Immediate", "Zoom") var aperture_change:int = ApertureChange.Immediate

@export var focal_length: float = 50.0:
    set(v):
        _dirty = true
        focal_length = v
        _new_fov = focal_length_to_fov(v)
@export var dof_far_range: float = 30.0:
    set(v):
        _dirty = true
        dof_far_range = v
        infinity_distance = v

        if $AFPoints:
            for ray in $AFPoints.get_children():
                ray.target_position.z = -dof_far_range
@export var autofocus: bool = false:
    set(v):
        _dirty = true
        _focused = null
        print("Changing autofocus to ", v)
        autofocus = v
@export var autofocus_threshold: float = 1.0
enum AutofocusMode {LOCK, CONTINUOUS, TARGET}
@export_enum("Lock", "Continous", "Selected Target") var autofocus_mode:int = AutofocusMode.CONTINUOUS
@export_node_path(Node3D) var autofocus_target_path:
    set(v):
        autofocus_target_path = v

@export var focus_speed: float = 1.0
@export_enum("Interpolate", "Const") var focus_accel:int = 0

@export var zoom_speed: float = 1.0

@export var blur_amount:float = 0.2:
    set(v):
        blur_amount = v
        _dirty = true
        _recreate_effect = true
@export var disabled: bool = false:
    set(v):
        disabled = v
        _dirty = true
        _on_change_disabled()
@export var dof_factor:float = 1.0:
    set(v):
        dof_factor = v
        _dirty = true
@export var disable_dof_far:bool = false:
    set(v):
        disable_dof_far = v
        _dirty = true
        _recreate_effect = true
@export var disable_dof_near:bool = false:
    set(v):
        disable_dof_near = v
        _dirty = true
        _recreate_effect = true
@export var focus_correction: float = 0.0:
    set(v):
        focus_correction = v
        _dirty = true
@export var focus_helper: bool = false:
    set(v):
        focus_helper = v
        _dirty = true



var dof_near_distance:float = 5.0
var dof_far_distance:float = 40.0
const unit = 1000.0  # 1 meter

const CoC = 0.03  # Full Frame
var near:float = 0.0
var far:float = 0.0
var dof:float = 0.0
var _fov:float = 0.0
var infinity_distance: float = dof_far_range
var target_distance = distance
var current_distance = distance
var _focused = null
var _aperture_changed = true
var _target_locked = false


var world_env: WorldEnvironment:
    set(x):
        world_env = x
        if world_env:
            env = x.environment
            effects = x.camera_effects
        else:
            if camera:
                env = camera.environment
                effects = camera.effects

var env: Environment
var effects: CameraEffects

# Called when the node enters the scene tree for the first time.


func _on_change_disabled():
    if effects:
        effects.dof_blur_near_enabled = not disabled
        effects.dof_blur_far_enabled = not disabled
        effects.override_exposure_enabled = not disabled

func focal_length_to_fov(fl):
    return 100 * atan(35.0 / fl)  # assumng 35mm full frame

func _ready():
    _fov = _new_fov  # initial fov == declared fov
    current_distance = distance
    camera = get_node(camera_path)
    if not camera:
        camera = get_parent()

    camera.fov = _fov

    _on_focus_checker_timeout()
    $FocusChecker.start()
    $AutofocusTimer.start()

    effects = world_env.effects if world_env else camera.effects
    env = world_env.environment if world_env else camera.environment

    _on_change_disabled()

func _calculate_dof(_dist):
    var hyperfocal = (focal_length * focal_length) / (aperture * CoC)
    near = ((hyperfocal * _dist) / (hyperfocal + (_dist - focal_length))) / unit
    far = ((hyperfocal * _dist) / (hyperfocal - (_dist - focal_length))) / unit
    dof = abs(far-near)

    near = (near - near * dof_factor) + current_distance + focus_correction
    far = (far + near * dof_factor) + current_distance + focus_correction
    return [dof, near, far]

# Called every frame. 'delta' is the elapsed time since the previous frame.
func _update(delta):
    if disabled:
        return

    if focus_helper and not $FocusHelper.visible:
        $FocusHelper.visible = true
    elif not focus_helper and $FocusHelper.visible:
        $FocusHelper.visible = false

    if autofocus:
        var spd = focus_speed * delta


        if focus_accel == 0:
            current_distance = lerp(current_distance, target_distance, focus_speed * delta * 0.5)
        else:
            current_distance = clampf(current_distance + sign(target_distance-current_distance) * spd, 0, max(target_distance, current_distance))

        #print_debug("cd=", current_distance, "  d=", distance)
        if not _focused and current_distance == target_distance:
            _focused = true
            if autofocus:
                emit_signal("autofocus_stop")
                if not _target_locked and autofocus_mode == AutofocusMode.LOCK:
                    _target_locked = true

            emit_signal("focused")
            if autofocus_mode == AutofocusMode.LOCK and target_distance < dof_far_range:
                _target_locked = true
    else:
        current_distance = target_distance

    if autofocus_mode == AutofocusMode.LOCK:
        if $LockThresholdTimer.is_stopped():
            $LockThresholdTimer.start()

    var _dist = current_distance * unit
    var _result = _calculate_dof(_dist)
    dof = _result[0]
    near = _result[1]
    far = _result[2]

    if far < 0 or far > infinity_distance:
        far = infinity_distance
        if effects.dof_blur_far_enabled:
            _recreate_effect = true
    elif not effects.dof_blur_far_enabled:
        _recreate_effect = true

    if _recreate_effect and effects:
        #effects = CameraEffects.new()
        effects.override_exposure_enabled = false
        effects.dof_blur_amount = blur_amount
        effects.dof_blur_far_transition = 5.0
        effects.dof_blur_near_transition = 5.0
        effects.dof_blur_far_enabled = not disable_dof_far and far > 0 and far < infinity_distance
        effects.dof_blur_near_enabled = not disable_dof_near
        _recreate_effect = false

    if effects:
        effects.dof_blur_far_distance = far + focus_correction
        effects.dof_blur_near_distance = near + focus_correction

    #print("current_distance=", current_distance)
    if focus_helper:
        $FocusHelper.position.z = -target_distance - $FocusHelper.radius
    #print_debug("camera.effects.dof_blur_far_enabled=", camera.effects.dof_blur_far_enabled)

    #camera.effects.override_exposure = -0.5 * pow(1.2, aperture) + 2.8
    if _aperture_changed and env:
        env.tonemap_exposure = 0.15 + sqrt(pow(2, -aperture)) * 4.0
        _aperture_changed = false

    if not _fov == _new_fov:
        if aperture_change == ApertureChange.Zoom:
            _fov = clampf(_fov + sign(_new_fov-_fov) * zoom_speed * delta, 0, max(_fov, _new_fov))
        else:
            _fov = _new_fov
        camera.fov = _fov

func _process(delta):
    if (autofocus and not _focused) or not _fov == _new_fov:
        _dirty = true

    if _dirty:
        _update(delta)
        _dirty = false


func _distance_to_focused_point():
    var target
    var selected_ray = $AFPoints/AutofocusRay

    if autofocus_target_path:
        # FIXME: not working well for objects of size near DOF or bigger
        # because of checking distance to it's origin, not face
        target = get_node(autofocus_target_path).global_transform.origin
    else:
        for ray in $AFPoints.get_children():
            if ray.is_colliding():
                # TODO: calculate weight and select most accurate ray coll point
                target = ray.get_collision_point()
                selected_ray = ray
                break

    var camp = camera.global_transform.origin
    if target:
        var actual_distance = camp.distance_to(target)
        return actual_distance
    else:
        return camp.distance_to(selected_ray.to_global(selected_ray.target_position))


func is_out_of_focus(distance_to_target):
    # >5% of DOF is out of focus
    #print_debug("dist: ",abs(distance_to_target-current_distance))
    #print_debug("far-near: ", abs(far-near)*1.1)

    #var _result = _calculate_dof(distance_to_target)
    var _dof = dof
    var _near = near
    var _far = far


    #return abs(distance_to_target-current_distance) > dof * autofocus_threshold * 0.1
    return distance_to_target < _near + 5.0 or distance_to_target > _far - 5.0

func _on_autofocus_timer_timeout():
    if not autofocus or (autofocus_mode == AutofocusMode.LOCK and _target_locked):
        return

    var _d = _distance_to_focused_point()
    if is_out_of_focus(_d):
        target_distance = _d  # constantly update target distance
        _dirty = true
        if _focused:
            _focused = false
            emit_signal("out_of_focus")
            emit_signal("autofocus_start")


func _on_focus_checker_timeout():
    if autofocus:
        return

    var _d = _distance_to_focused_point()
    var _nf = not is_out_of_focus(_d)
    if not _nf == _focused:
        emit_signal("out_of_focus" if not _nf else "focused")
        _focused = _nf

func _on_lock_threshold_timer_timeout():
    if _target_locked and is_out_of_focus(_distance_to_focused_point()):
        _target_locked = false
